;
; channel15_2.asm - channel 15 command execution
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;

channel15_cont	.proc

		ldx #commandslo-commands
htnuo		cmp commands-1,x
		bne njooocmd
		lda commandshi-1,x
		pha
		lda commandslo-1,x
		pha
		lda path_exp+1
		rts

njooocmd	dex
		bne htnuo

cmdnotimp	.proc
		sec
		gcs LWOP_reporterror_31_xit;Unknown command
		.pend
                
cmdsynterr	.proc
		sec
		gcs LWOP_reporterror_30_xit;Command syntax
		.pend

;-------------------------------
;U
;
cmd_user	.proc
		cmp #"0"
		beq usercmd0
		and #15
		cmp #10
		beq usercmdreset
		cmp #9
		bne cmdnotimp
;-------------------------------
;UI/U9
;
usercmd9	.proc
		ldx #0
		ldy #0
		jmp LWOP_reporterror_73_xit
		.pend
;-------------------------------
;UJ/U:
;
usercmdreset	.proc
;       	lda current_drive_exp	;no detection, cannot reset only _one_ of the devices
;		jsr OP_hdinit
		lda current_drive_exp
		jmp OP_XIT_hdinit_2
		.pend
;-------------------------------
;U0>
;
usercmd0	.proc
		lda path_exp+2
		cmp #">"
		bne cmdsynterr
		lda path_exp+3
in		cmp #31			;"U0>"+CHR$(<drivenumber>)
		bcs usercmdnnumber
		#ldx_l_ current_drive_exp
		cmp #0			;Disable drive
		beq tnstns
		sta current_drivenumbers_exp,x
		jmp LWOP_reporterror_ok
tnstns		sta drive_type_exp,x
		geq cmdusercmddisab	;Power down disabled drive
		.pend
;-------------------------------
;U0>P
;
usercmdnnumber	.proc
		cmp #"P"		;"U0>P"+CHR$(<powermode>)
		bne usercmdnp
		jsr head_select_drive
		lda path_exp+4
		lsr
		bcc cmdusercmddisab
		lda #idle_imediate1_d
		jsr ata_command
		bcc cmdusercmddisab.ok
		lda #idle_imediate2_d
		gcs cmdusercmddisab.be
cmdusercmddisab	.proc
		lda #standby_now1_d
		jsr ata_command
		bcc ok
		lda #standby_now2_d
be		jsr ata_command
		jsr LWOP_reporterror_xit
ok		jmp LWOP_reporterror_ok
		.pend

;-------------------------------
;U0>E
;
usercmdnp	.proc
		cmp #"E"		;"U0>E"+CHR$(<mode>)
		bne usercmdne
		jsr XIT_init_drive2
		lda #ATAPI_CMD_START_STOP
		jsr OP_atapi_packet
		lda path_exp+4
		and #1
		eor #3
                gne usercmdne.in
;-------------------------------
;U0>L
;
usercmdne	.proc
		cmp #"L"		;"U0>L"+CHR$(<mode>)
		bne usercmdnl
		jsr XIT_init_drive2
		lda #ATAPI_CMD_LOCK_UNLOCK
		jsr OP_atapi_packet
		lda path_exp+4
		and #1
in		#pw 4
		jsr OP_atapi_command
		jsr LWOP_reporterror_xit
		jmp LWOP_reporterror_ok
		.pend
;-------------------------------
;U0>A
;
usercmdnl	.proc
		cmp #"A"		;"U0>A"+CHR$(M)+CHR$(S)+CHR$(F)+CHR$(M)+CHR$(S)+CHR$(F)
		bne cmdnotimp
		jsr XIT_init_drive2
		lda #ATAPI_PLAY_AUDIO_MSF
		jsr OP_atapi_packet
		lda path_exp+4
		#pw 3
		lda path_exp+5
		#pw 4
		lda path_exp+6
		#pw 5
		lda path_exp+7
		#pw 6
		lda path_exp+8
		#pw 7
		lda path_exp+9
		#pw 8
		jsr OP_atapi_command
		jsr LWOP_reporterror_xit
		jmp LWOP_reporterror_ok
		.pend
;-------------------------------
;G get partition
;
;"G-P"           Current partition
;"G-P"+CHR$(255) Current partition
;"G-P"+CHR$(k)   Other partition
;
;1st byte: 0 not created, 1 native, 2 1541, 3 1571, 4 1581, 5 1581 cpm, 6 Spool, 7 Foreign, 255 system
;2nd byte: info
;3rd byte: partition
;4-19th byte: Partition name
;20-28th byte: ???
;29-30th byte: size of partition

cmd_g		.proc
		jsr OP_short_path
                cmp #"-"
                bne nkq
		jsr OP_short_path
                cmp #"P"
                bne nkq
                jsr OP_short_path
                cmp #255
                bne *+4
                lda #0
                pha

		jsr XIT_init_drive	;-=> error

                pla
                jsr OP_check_partition_selection
                #lda0
                bcs tt
                iny
                lda #1
tt              pha

		#ldatax current_drive_exp
                lda #3+16+9+2
                sta status_size_exp,x
                #lda0
                sta status_pos_exp,x
		txa
		lsr
		ror			;*64
		ror
		tax
                pla			;Native partition
                sta drive_status_exp,x
                lda #$e0		;Disk in, formatted, CMD formatted, unknown size
                sta drive_status_exp+1,x
                tya
                sta drive_status_exp+2,x
                lda #32
                ldy #16+9
nk              sta drive_status_exp+3,x
                inx
                dey
                bne nk
                lda #255
                sta drive_status_exp+3,x
                sta drive_status_exp+4,x
		rts

nkq		jmp cmdnotimp
		.pend
;-------------------------------
;S remove file
;
;"S-8" swap 8
;"S-9" swap 9
;"S-D" swap default
;
cmd_s		.proc
                jsr OP_short_path
                cmp #"8"
                beq nj2
                cmp #"9"
                beq nj2
                cmp #"D"
                beq nj3
		sec
		gcs LWOP_reporterror_31_xit;Unknown command
nj2		and #$0f
ej		#jmps usercmd0.in

nj3		#ldatax current_drive_exp
                lda setup_drivenumbers_exp,x
                and #$0f
                clc
                adc #7
                gcc ej
                .pend
;-------------------------------
;Fake memory read and write
;
;fea0 "CMD HD" CMD HD
;fea0 "CMD RL" CMD RamLink
;fea0 "CMD RD" CMD RamDrive
;e5c6 "4"      1541
;e5c6 "7"      1571
;a6e9 "8"      1581
;fea0 "CMD FD" CMD FD
;
cmd_memory      .proc
		jsr OP_short_path
                cmp #"W"
                beq cmd_skip
                cmp #"R"
                bne cmdnotimp
		jsr OP_short_path
                pha
		jsr OP_short_path
		jsr OP_short_path
                cmp #60
                bcc *+4
                lda #60
                sta pom1
                
                #ldx_l_ current_drive_exp
                sta status_size_exp,x
                #lda0
                sta status_pos_exp,x
		txa
		lsr
		ror			;*64
		ror
		tax
                #ply_a
                lda setup_atribut2_exp	;CMD emulation?
                asl

lul             tya
		and #15
                bcs *+4
                ora #16
                tay
		lda ou,y
                sta drive_status_exp,x
                inx
                iny
                dec pom1
                bpl lul
		rts

ou              .text "CMD HD EMULATED "	;Fake drive ROM
ou2             .text "IDE64 CARTRIDGE "	;Fake drive ROM
		.pend
                
cmd_skip        .proc
		clc
		gcc LWOP_reporterror_ok
		.pend
;-------------------------------
;T Time
;
;T-RA
;T-RB
;T-RD
;T-WA
;T-WB
;T-WD
;
;SUN. 00/00/00 00:00:00 AM
;
cmd_t		.proc
                jsr OP_short_path
                cmp #"-"
                bne nkq
                jsr OP_short_path
                cmp #"R"
                bne nkq
                jsr OP_short_path
                cmp #"A"
                beq q
                cmp #"B"
                beq q
                cmp #"D"
                bne nkq
q		pha
                
		jsr LWOP_read_clock

;00 -> 12 AM
;01 -> 01 AM
;..
;11 -> 11 AM
;12 -> 12 PM
;13 -> 01 PM
;..
;23 -> 11 PM
                lda hour_exp
                cmp #$12
                php
                tax
                bne *+4
                lda #$12
                cmp #$13
                bcc xq
                php
                sei
                sed
                sbc #$12
                plp
xq              plp
                rol
                sta hour_exp

		pla
                cmp #"D"
                bne nu

                ldy #6
rk
		.if SUPERCPU=1
                tyx
                lda sec_exp,x
                .else
                lda sec_exp,y
                .fi
                jsr bcd2bin
                .if SUPERCPU=1
                tyx
                sta sec_exp,x
                .else
                sta sec_exp,y
                .fi
                dey
                bpl rk
                tya
nu
		cmp #"B"

                #ldatax current_drive_exp
		lda #8
                bcs *+4
                lda #25
                sta status_size_exp,x
                php
                #lda0
                sta status_pos_exp,x
		txa
		lsr
		ror			;*64
		ror
		tax

                lda day_exp
                #sbc1c
                plp
                bcc asc
                sta drive_status_exp,x
                lda year_exp
                sta drive_status_exp+1,x
                lda month_exp
                sta drive_status_exp+2,x
                lda date_exp
                sta drive_status_exp+3,x
                lda hour_exp
                lsr
                php
		sta drive_status_exp+4,x
                lda min_exp
                sta drive_status_exp+5,x
                lda sec_exp
                sta drive_status_exp+6,x
                plp
                #lda0
                rol
                sta drive_status_exp+7,x
		rts
                
asc		asl
		asl
                tay
nuo             lda days,y
                sta drive_status_exp,x
                iny
                inx
                txa
                and #3
                bne nuo
                ldy #$20
                lda month_exp
                jsr appd
                ldy #"/"
                lda date_exp
                jsr appd
                lda year_exp
                jsr appd
                ldy #$20
                lda hour_exp
                lsr
                php
                jsr appd
                ldy #":"
                lda min_exp
                jsr appd
                lda sec_exp
                jsr appd
                lda #$20
                jsr aaa
                plp
                lda #"A"
                bcc *+4
                lda #"P"
                jsr aaa
                lda #"M"
                gne aaa

appd            pha
		tya
                jsr aaa
                pla
                pha
		lsr
                lsr
                lsr
                lsr
                jsr aa
                pla
aa		and #$0f
                ora #$30
aaa		sta drive_status_exp,x
                inx
		rts

nkq		jmp cmdnotimp

days		.text "SUN."
		.text "MON."
		.text "TUES"
		.text "WED."
		.text "THUR"
		.text "FRI."
		.text "SAT."
                .pend
;-------------------------------
;N format
;
;
cmd_n		.proc
                jsr OP_short_path
                cmp #"="
                bne nkq
                jsr OP_short_path
                
                ldy #0
                clc
ne2             ldx #0
ne              lda formats,y
                bmi ni
                eor path_exp,x
                beq *+3
                sec
                iny
                inx
                gne ne

ni		bcs ee
		and #$7f
		eor path_exp,x
                beq joo
ee              tya
                clc
                adc #4
                tay
                cpy #formatsend-formats
                bcc ne2
nkq		jmp cmdnotimp

joo		#phy_a
		jsr XIT_init_drive2

		lda #ATAPI_FORMAT_UNIT
		jsr OP_atapi_packet
		#pc 1,%00010111
		jsr OP_atapi_dataout_command

		#lda0
		jsr OP_atapi_packet

             	#pc 1,%10100010
		#pc 3,8
		#pc 10,>512
                
                #ply_a                
                ldx #2
lu              iny
		lda formats,y
                sta packet_buff_exp+4,x
                dex
                bpl lu

		ldx #11
lp		lda packet_buff_exp,x	;Send the command to the device.
		sta low_reg
		dex
		lda packet_buff_exp,x
		sta hig_reg
		sta data_g
		dex
		bpl lp
                jsr ata_pi_nomore_data
		bcs nj
		jsr wait_busy		;wait until writing finished
nj              jsr LWOP_reporterror_write_xit
		jsr invalidate_drive		
		jmp LWOP_reporterror_ok
                
formats         .text "720","K"+128
                .byte $00,$05,$a0	;80/2/9 75Kbps 720rpm
                .text "1.2","M"+128
                .byte $00,$09,$60	;80/2/15 125Kbps 720rpm
                .text "1.44","M"+128
                .byte $00,$0b,$40	;80/2/18 150Kbps 720rpm
                .text "120","M"+128
                .byte $03,$c3,$00	;963/8/32 533Kbps 720rpm
;                .text "100","M"+128
;                .byte $00,$02,$FF,$E0	;96/64/32 4096Kbps 2941rpm (does not work)
formatsend
                .pend

commands	.text "GUSMVNT"
commandslo    	.byte <(cmd_g-1)
		.byte <(cmd_user-1)
		.byte <(cmd_s-1)
		.byte <(cmd_memory-1)
		.byte <(cmd_skip-1)
		.byte <(cmd_n-1)
		.byte <(cmd_t-1)
commandshi     	.byte >(cmd_g-1)
		.byte >(cmd_user-1)
		.byte >(cmd_s-1)
		.byte >(cmd_memory-1)
		.byte >(cmd_skip-1)
		.byte >(cmd_n-1)
		.byte >(cmd_t-1)
